<?php

namespace Drupal\component_schema\Plugin\Styleguide;

use Drupal\component_schema\Component\TypedComponentManager;
use Drupal\styleguide\Plugin\StyleguidePluginBase;
use Drupal\Core\Extension\Extension;
use Drupal\Core\Extension\InfoParserInterface;
use Drupal\Core\Extension\ModuleExtensionList;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Extension\ThemeExtensionList;
use Drupal\Core\Url;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Component Schema items implementation.
 *
 * @Plugin(
 *   id = "component_schema",
 *   label = @Translation("Component Schema elements")
 * )
 */
class ComponentSchema extends StyleguidePluginBase {

  /**
   * The module handler service.
   *
   * @var ModuleHandlerInterface
   */
  protected $moduleHandler;

  /**
   * The module extension list.
   *
   * @var \Drupal\Core\Extension\ModuleExtensionList
   */
  protected $moduleExtensionList;

  /**
   * The theme extension list.
   *
   * @var \Drupal\Core\Extension\ThemeExtensionList
   */
  protected $themeExtensionList;

  /**
   * The info parser.
   *
   * @var \Drupal\Core\Extension\InfoParserInterface
   */
  protected $infoParser;

  /**
   * The typed component manager.
   *
   * @var \Drupal\component_schema\Component\TypedComponentManager
   */
  protected $typedComponentManager;

  /**
   * Constructs a new ComponentSchema.
   *
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin ID for the plugin instance.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\Core\Extension\ModuleHandlerInterface $module_handler
   *   The module handler
   * @param \Drupal\Core\Extension\ModuleExtensionList $extension_list_module
   *   The module extension list
   * @param \Drupal\Core\Extension\ThemeExtensionList $extension_list_theme
   *   The theme extension list
   * @param \Drupal\Core\Extension\InfoParserInterface $info_parser
   *   The info parser.
   * @param \Drupal\component_schema\Component\TypedComponentManager $typed_component_manager
   *   The typed component manager.
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition, ModuleHandlerInterface $module_handler, ModuleExtensionList $extension_list_module, ThemeExtensionList $extension_list_theme, InfoParserInterface $info_parser, TypedComponentManager $typed_component_manager) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);

    $this->moduleHandler = $module_handler;
    $this->moduleExtensionList = $extension_list_module;
    $this->themeExtensionList = $extension_list_theme;
    $this->infoParser = $info_parser;
    $this->typedComponentManager = $typed_component_manager;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('module_handler'),
      $container->get('extension.list.module'),
      $container->get('extension.list.theme'),
      $container->get('info_parser'),
      $container->get('component_schema.typed')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function items() {
    $items = [];

    foreach ($this->typedComponentManager->getDefinitions() as $machine_name => $component) {
      // The definitions include base types. We only want the component types.
      if (empty($component['type']) || ($component['type'] !== 'component_type_mapping')) {
        continue;
      }
      switch ($component['extension_type']) {
        case 'module':
          $extension = $this->moduleExtensionList->get($component['extension_name']);
          break;
        case 'theme':
          $extension = $this->themeExtensionList->get($component['extension_name']);
          break;
      }
      $extension_info = $this->getExtensionInfo($extension);
      $description = [
        '#type' => 'details',
        '#title' => $this->t('Component details'),
        '#open' => FALSE,
      ];
      if (isset($component['settings']['content_class'])) {
        $description['#attributes'] = [
          'class' => [$component['settings']['content_class']],
        ];
      }
      $description_items = [];
      $description_items['description'] = [
        '#type' => 'html_tag',
        '#tag' => 'p',
        '#value' => $this->t($component['description']),
      ];
      $description_items['template'] = [
        '#type' => 'html_tag',
        '#tag' => 'p',
        '#value' => $this->t('Component template: <code>@component_template</code>', ['@component_template' => $component['component_template']]),
      ];
      // If data from drupal.org is present, assume this is a contributed
      // extension and link to it on drupal.org.
      if (isset($extension_info['project']) && isset($extension_info['version']) && isset($extension_info['datestamp'])) {
        $description_items['provider'] = [
          '#type' => 'html_tag',
          '#tag' => 'p',
          '#value' => $this->t('Provided by the @type <strong><a href=":project">@title</a></strong> (<code>@name</code>)', ['@type' => $this->t($component['extension_type']), ':project' => 'https://www.drupal.org/project/' . $extension_info['project'], '@title' => $extension_info['name'], '@name' => $component['extension_name']]),
        ];
      }
      // Otherwise, just give the extension name.
      else {
        $description_items['provider'] = [
          '#type' => 'html_tag',
          '#tag' => 'p',
          '#value' => $this->t('Provided by the @type <strong>@title</strong> (<code>@name</code>)', ['@type' => $this->t($component['extension_type']), '@title' => $extension_info['name'], '@name' => $component['extension_name']]),
        ];
      }

      if (!empty($component['documentation_url'])) {
        $description_items['documentation'] = [
          '#type' => 'link',
          '#title' => $this->t('View component documentation'),
          '#url' => Url::fromUri($component['documentation_url']),
          '#attributes' => ['target' => '_blank'],
        ];
      }
      $description['basics'] = [
        '#theme' => 'item_list',
        '#items' => $description_items,
        '#title' => $this->t('Basics'),
      ];
      if (!empty($component['mapping'])) {
        $description['mapping'] = $this->getVariablesItemList($component['mapping']);
      }
      // Prefix the key to minimize collisions with items provided by other
      // modules.
      $items['component_' . $machine_name] = [
        'title' => $this->t($component['label']),
        'description' => $description,
        'content' => [
          '#type' => 'inline_template',
          '#template' => '{% include(stylesheet_template) %}',
          '#context' => [
            'stylesheet_template' => $component['stylesheet_template'],
            'base_path' => base_path(),
            'extension_path' => $extension->getPath(),
          ],
        ],
        'group' => $this->t($component['group']),
      ];
    }

    return $items;
  }

  /**
   * Gets a nested #item_list element for variables.
   *
   * @param array $variables
   *   Variables to pass to a template.
   * @param bool $use_title
   *   Used internally. Whether or not to use the title "Variables".
   *
   * @return array
   *   A render array of variables suitable for passing to theme(item_list).
   */
  protected function getVariablesItemList(array $variables, $use_title = TRUE) {
    $items = [];
    foreach ($variables as $name => $variable) {
      $item = [];

      // Convert types as needed. We're repurposing types from Drupal core's
      // configuration schema. Substitute type names suitable to Twig
      // templates.
      // @see https://www.drupal.org/docs/8/api/configuration-api/configuration-schemametadata
      switch ($variable['type']) {
        case 'mapping':
        case 'component_mapping':
        case 'component_type_mapping':
          $type = 'hash';
          break;
        case 'sequence':
        case 'component_sequence':
          $type = 'array';
          break;
        case 'component_string':
          $type = 'string';
          break;
        case 'component_boolean':
        case 'boolean':
          $type = 'bool';
          break;
        case 'component_attribute':
          $type = '\Drupal\Core\Template\Attribute';
          break;
        case 'ignore':
          $type = 'mixed';
          break;
        default:
          $type = $variable['type'];
      }

      $default = '';

      $options = [
        '@name' => $name,
        '@type' => $type,
        '@label' => $this->t($variable['label'] ?? ''),
      ];
      if (!empty($variable['default'])) {
        $options['%default'] = $variable['default'];
      }
      // Add a description of the variable.
      $item['description'] = [
        '#type' => 'markup',
      ];
      // Vary the output according to whether the variable is optional
      // (nullable) and whether there's a default value.
      if (empty($variable['nullable'])) {
        if (empty($variable['default'])) {
          $item['description']['#markup'] = $this->t('<code>@name</code> (<code>@type</code>): @label', $options);
        }
        else {
          $item['description']['#markup'] = $this->t('<code>@name</code> (<code>@type</code>, defaults to %default): @label', $options);
        }
      }
      else {
        if (empty($variable['default'])) {
          $item['description']['#markup'] = $this->t('<code>@name</code> (<code>@type</code>, optional): @label', $options);
        }
        else {
          $item['description']['#markup'] = $this->t('<code>@name</code> (<code>@type</code>, optional, defaults to %default): @label', $options);
        }
      }

      // Link to the styleguide documentation.
      if (!empty($variable['component_type'])) {
        $item['link'] = [
          '#prefix' => ' ',
          '#type' => 'markup',
          '#markup' => $this->t('<a href="#component_@type">View component</a>', ['@type' => $variable['component_type']]),
        ];
      }

      // Add an optional documentation URL. This property is a custom addition
      // to the schema format used in Drupal core.
      elseif (!empty($variable['documentation_url'])) {
        $item['link'] = [
          '#prefix' => ' » ',
          '#type' => 'link',
          '#title' => $this->t('View variable documentation'),
          '#url' => Url::fromUri($variable['documentation_url']),
          '#attributes' => ['target' => '_blank'],
        ];
      }

      // Iterate through any child variables.
      switch ($variable['type']) {
        case 'mapping':
        case 'component_mapping':
        case 'component_type_mapping':
          $item['children'] = $this->getVariablesItemList($variable['mapping'], FALSE);
          break;
        case 'sequence':
        case 'component_sequence':
          $item['sequence'] = [
            '#prefix' => ' » ',
            '#type' => 'markup',
            '#markup' => $this->t('An array of items, each with the following variables:'),
          ];
          if (isset($variable['sequence']['type']) && in_array($variable['sequence']['type'], ['mapping', 'component_mapping', 'component_type_mapping'])) {
            $item['children'] = $this->getVariablesItemList($variable['sequence']['mapping'], FALSE);
          }
          break;
      }

      $items[] = $item;
    }

    $variables = [
      '#theme' => 'item_list',
      '#items' => $items,
    ];

    if ($use_title) {
      $variables['#title'] = $this->t('Variables');
    }

    return $variables;
  }

  /**
   * Gets the info for a given module or theme extension.
   *
   * @param \Drupal\Core\Extension\Extension $extension
   *   A module or theme extension.
   *
   * @return array
   *   An array of information parsed from the extension's .info.yml file.
   */
  protected function getExtensionInfo(Extension $extension) {
    static $info = [];

    $type = $extension->getType();
    $name = $extension->getName();
    if (!isset($info[$type])) {
      $info[$type] = [];
    }
    if (!isset($info[$type][$name])) {
      $extension_info = $this->infoParser->parse($extension->getPathname());
      // Invoke hook_system_info_alter() to give installed modules a chance to
      // modify the data in the .info.yml files if necessary.
      $this->moduleHandler->alter('system_info', $extension_info, $extension, $type);
      $info[$type][$name] = $extension_info;
    }

    return $info[$type][$name];
  }

}
